#!/usr/bin/env python3

from __future__ import print_function
import argparse
import io
import logging
import os
import re
import sys
from copy import deepcopy
from jinja2 import Environment, FileSystemLoader, StrictUndefined
from jinja2.exceptions import TemplateNotFound
from ruamel.yaml import YAML


def parse_template(yaml_string):
    """
    Round trip lava_job through ruamel to test parsing and improve formatting.
    Comments are preserved.

    In: yaml-formatted string
    Out: validated yaml-formatted string
    """
    yaml = YAML()
    yaml.indent(sequence=4, offset=2)
    # ruamel does not provide a mechanism to dump to string, so use StringIO to catch it
    output = io.StringIO()
    yaml.dump(yaml.load(yaml_string), output)
    return output.getvalue()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--template",
        help="""A Jinja2 template for the yml file to be genarated.
                All variables in the templates will be substituted using the
                Jinja2 engine. Variables can be passed either as environment 
                variables or as command line arguments
                (e.g. --my-variable=myvalue). In the latter, hyphens ("-") are 
                replaced with underscores ("_"), i.e. "--my-variable" 
                becomes "my_variable".""",
        required=True,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="""Increase verbosity.""",
    )
    parser.add_argument(
        "arguments",
        nargs="*",
        help="""Arguments passed as list to the template.""",
    )

    # parse known args
    args, unknown = parser.parse_known_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    logging.debug("Using template: %s", args.template)

    # parse optional arguments
    parser = argparse.ArgumentParser()
    for arg in unknown:
        if arg.startswith(("-", "--")):
            parser.add_argument(arg.split("=")[0])
    optional_args, _ = parser.parse_known_args()
    logging.debug("Optional args: %s", optional_args)

    # update context with given optional arguments
    j2_env = Environment(undefined=StrictUndefined, extensions=["jinja2.ext.do"])
    context = dict(deepcopy(os.environ))
    for arg in vars(optional_args):
        context[arg] = vars(optional_args)[arg]
    if args.arguments:
        context["optional_arguments"] = args.arguments

    for k, v in context.items():
        logging.debug("Optional args: %s -- %s", k, v)

    # render the template
    j2_env.loader = FileSystemLoader(os.path.dirname(args.template))

    try:
        lava_job = j2_env.get_template(os.path.basename(args.template)).render(context)
        lava_job = parse_template(lava_job)
        # Remove duplicate newlines
        lava_job = re.sub(r"\n+", r"\n", lava_job)

        print(lava_job)

    except TemplateNotFound:
        print("ERROR: Template %s not found" % args.template, file=sys.stderr)
        exit(1)


if __name__ == "__main__":
    main()
