#!/usr/bin/env python3
import common

import argparse
import sys
from gitlab import Gitlab, GitlabGetError


def get_merge_requests(
    project, state=None, source_branch=None, target_branch=None, commit=None
):
    """Get merge request by source and target branch and optionally commit sha"""
    merge_requests = []
    try:
        all_merge_requests = project.mergerequests.list(
            source_branch=source_branch,
            target_branch=target_branch,
            state=state if state else "all",
            all=True,
            retry_transient_errors=True,
        )
    except GitlabGetError as e:
        sys.exit(
            "ERROR: could not list merge requests for project '%s': %s"
            % (project.name, e)
        )
    if commit:
        for mr in all_merge_requests:
            if mr.sha == commit or mr.squash_commit_sha == commit:
                merge_requests.append(mr)
    elif all_merge_requests:
        merge_requests = all_merge_requests

    # Get complete objects
    full_merge_requests = []
    for mr in merge_requests:
        mr = project.mergerequests.get(mr.iid, retry_transient_errors=True)
        full_merge_requests.append(mr)

    return full_merge_requests


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        required=True,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        help="""name of the GitLab project""",
        dest="project",
        required=True,
    )
    parser.add_argument(
        "--state",
        help="""state of the merge request (opened, closed, locked, or merged)""",
        dest="state",
        required=True,
    )
    parser.add_argument(
        "--source-branch",
        help="""source branch of the merge request""",
        dest="source_branch",
        required=False,
    )
    parser.add_argument(
        "--target-branch",
        help="""target branch of the merge request""",
        dest="target_branch",
        required=False,
    )
    parser.add_argument(
        "--commit",
        help="""commit sha of the merge request""",
        dest="commit",
        required=False,
    )

    args, _ = parser.parse_known_args()

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)
    project = common.get_project(gitlab, args.project)

    mrs = get_merge_requests(
        project,
        state=args.state,
        source_branch=args.source_branch,
        target_branch=args.target_branch,
        commit=args.commit,
    )

    for mr in mrs:
        print(mr.iid)


if __name__ == "__main__":
    main()
