#!/usr/bin/env python3
import common

import argparse
import sys
from gitlab import Gitlab, GitlabGetError
from gitlab.v4.objects import Project, ProjectPipeline
from typing import List


def retrigger_pipeline_jobs(
    project: Project,
    pipeline: ProjectPipeline,
    job_name: str,
    status_list: List[str],
    include_children: bool = False,
):
    jobs = []

    # Recurse to child pipelines
    if include_children:
        for bridge in pipeline.bridges.list():
            if (
                bridge is not None
                and bridge.downstream_pipeline is not None
                and bridge.downstream_pipeline["project_id"] == project.id
            ):
                child_pipeline = project.pipelines.get(
                    bridge.downstream_pipeline["id"],
                    retry_transient_errors=True,
                )
                jobs += retrigger_pipeline_jobs(
                    project,
                    child_pipeline,
                    job_name,
                    status_list,
                    include_children=True,
                )

    # Find job
    job = None
    for pipelinejob in pipeline.jobs.list():
        if pipelinejob.name == job_name:
            job = project.jobs.get(pipelinejob.id, retry_transient_errors=True)
    if not job:
        return jobs

    # Only retrigger if job is in certain status
    if job.status not in status_list:
        return jobs

    # Retrigger job
    job.retry()
    print("Retrigger job '%s' in %s:" % (job_name, pipeline.web_url))
    job = project.jobs.get(job.id, retry_transient_errors=True)
    print(job.web_url)
    jobs.append(job)

    return jobs


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--gitlab-url",
        help="""URL to the GitLab instance""",
        dest="gitlab_url",
        required=True,
    )
    parser.add_argument(
        "--token",
        help="""GitLab REST API private access token""",
        dest="token",
        required=True,
    )
    parser.add_argument(
        "--project",
        help="""name of the GitLab project""",
        dest="project",
        required=True,
    )
    parser.add_argument(
        "--pipeline",
        help="""pipeline id""",
        dest="pipeline",
        required=True,
    )
    parser.add_argument(
        "--job",
        help="""name of the job""",
        dest="job",
        required=True,
    )

    parser.add_argument(
        "--include-children",
        help="""search for job in child pipelines""",
        dest="include_children",
        action="store_true",
        default=False,
    )

    args, _ = parser.parse_known_args()

    gitlab = Gitlab(args.gitlab_url, private_token=args.token)
    project = common.get_project(gitlab, args.project)
    try:
        pipeline = project.pipelines.get(args.pipeline, retry_transient_errors=True)
    except GitlabGetError as e:
        sys.exit("ERROR: could not get pipeline %s: %s" % (args.pipeline, e))
    if not pipeline:
        sys.exit("ERROR: could not find pipeline %s" % args.pipeline)

    jobs = retrigger_pipeline_jobs(
        project,
        pipeline,
        args.job,
        ["success", "running"],
        args.include_children,
    )

    print("Retriggered %d jobs for pipeline #%s" % (len(jobs), args.pipeline))


if __name__ == "__main__":
    main()
